from dataclasses import dataclass

import chevron
import logging
import re
import sass
import secrets
from base64 import urlsafe_b64encode
from chevron import ChevronError
from misaka import HtmlRenderer, Markdown
from typing import Dict

markdown = Markdown(HtmlRenderer(flags=['hard-wrap']), extensions=['no-intra-emphasis'])
logger = logging.getLogger('render')


@dataclass
class MessageDef:
    first_name: str
    last_name: str
    main_template: str
    mustache_partials: Dict[str, dict]
    macros: Dict[str, dict]
    subject_template: str
    context: dict
    headers: dict


@dataclass
class EmailInfo:
    full_name: str
    subject: str
    html_body: str
    headers: dict
    shortened_link: list


def _update_context(context, partials, macros):
    for k, v in context.items():
        if k.endswith('__md'):
            yield k[:-4], markdown(v)
        elif k.endswith('__sass'):
            yield k[:-6], sass.compile(string=v, output_style='compressed', precision=10).strip('\n')
        elif k.endswith('__render'):
            v = chevron.render(_apply_macros(v, macros), data=context, partials_dict=partials)
            yield k[:-8], markdown(v)


def _apply_macros(s, macros):
    if macros:
        for key, body in macros.items():
            m = re.search(r'^(\S+)\((.*)\) *$', key)
            if not m:
                logger.warning('invalid macro "%s", skipping it', key)
                continue
            name, arg_defs = m.groups()
            arg_defs = [a.strip(' ') for a in arg_defs.split('|') if a.strip(' ')]

            def replace_macro(m):
                arg_values = [a.strip(' ') for a in m.groups()[0].split('|') if a.strip(' ')]
                if len(arg_defs) != len(arg_values):
                    logger.warning('invalid macro call "%s", not replacing', m.group())
                    return m.group()
                else:
                    return chevron.render(body, data=dict(zip(arg_defs, arg_values)))

            s = re.sub(r'%s\((.*?)\)' % name, replace_macro, s)
    return s


SKIPPED_LINKS = [
    re.compile(r'\.(?:png|jpg|bmp)$'),
    re.compile(r'^https?://maps.googleapis.com'),
    re.compile(r'^https?://maps.google.com'),
]


def looks_like_link(s):
    return isinstance(s, str) and re.match('^https?://', s) and not any(m.search(s) for m in SKIPPED_LINKS)


def apply_short_links(context, click_url, click_random=30, backup_arg=False):
    shortened_link = []
    extra = {}
    for k, v in context.items():
        # TODO deal with unsubscribe links properly
        if k != 'unsubscribe_link' and looks_like_link(v):
            r = secrets.token_urlsafe(click_random)[:click_random]
            new_url = click_url + r
            if backup_arg:
                new_url += '?u=' + urlsafe_b64encode(v.encode()).decode()
            extra[k] = new_url
            extra[f'{k}_original'] = v
            shortened_link.append((v, r))
    context.update(extra)
    return shortened_link


def render_email(m: MessageDef, click_url=None, click_random=30) -> EmailInfo:
    full_name = f'{m.first_name or ""} {m.last_name or ""}'.strip(' ')
    m.context.setdefault('recipient_name', full_name)
    m.context.setdefault('recipient_first_name', m.first_name or full_name)
    m.context.setdefault('recipient_last_name', m.last_name)
    try:
        subject = chevron.render(m.subject_template, data=m.context)
    except ChevronError as e:
        logger.warning('invalid subject template: %s', e)
        subject = m.subject_template

    shortened_link = []
    if click_url:
        shortened_link = apply_short_links(m.context, click_url, click_random, backup_arg=True)
    m.context.update(email_subject=subject, **dict(_update_context(m.context, m.mustache_partials, m.macros)))
    unsubscribe_link = m.context.get('unsubscribe_link')
    if unsubscribe_link:
        m.headers.setdefault('List-Unsubscribe', f'<{unsubscribe_link}>')

    return EmailInfo(
        full_name=full_name,
        subject=subject,
        html_body=chevron.render(
            _apply_macros(m.main_template, m.macros), data=m.context, partials_dict=m.mustache_partials
        ),
        headers=m.headers,
        shortened_link=shortened_link,
    )


BASIC_CHARACTERS = {
    # basic characters from https://support.messagebird.com/hc/en-us/articles/208739765-Which-special-
    # characters-count-as-two-characters-in-a-text-message- ordered and repeats removed!
    ' ',
    '!',
    '"',
    '#',
    '$',
    '%',
    '&',
    "'",
    '(',
    ')',
    '*',
    '+',
    ',',
    '-',
    '.',
    '/',
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    ':',
    ';',
    '<',
    '=',
    '>',
    '?',
    '@',
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'O',
    'P',
    'Q',
    'R',
    'S',
    'T',
    'U',
    'V',
    'W',
    'X',
    'Y',
    'Z',
    '_',
    'a',
    'b',
    'c',
    'd',
    'e',
    'f',
    'g',
    'h',
    'i',
    'j',
    'k',
    'l',
    'm',
    'n',
    'o',
    'p',
    'q',
    'r',
    's',
    't',
    'u',
    'v',
    'w',
    'x',
    'y',
    'z',
    '¡',
    '£',
    '¤',
    '¥',
    '§',
    '¿',
    'Ä',
    'Å',
    'Æ',
    'Ç',
    'É',
    'Ñ',
    'Ö',
    'Ø',
    'Ü',
    'ß',
    'à',
    'ä',
    'å',
    'æ',
    'è',
    'é',
    'ì',
    'ñ',
    'ò',
    'ö',
    'ø',
    'ù',
    'ü',
    'Γ',
    'Δ',
    'Θ',
    'Λ',
    'Ξ',
    'Π',
    'Σ',
    'Φ',
    'Ψ',
    'Ω',
    # special cases from https://support.messagebird.com/hc/en-gb/articles/200731072-In-which-charset-can-I-deliver-
    # SMS-messages-and-what-should-I-take-into-consideration- this is not an exhaustive list :(
    'ç',
    '®',
}

# from first link above
# apparently messagebird call \n an extension character which differs from https://en.wikipedia.org/wiki/GSM_03.38
# this might be because they replace \n with LF + CR which would constitute 2 characters
EXTENSION_CHARACTERS = {'\n', '[', '\\', ']', '^', '{', '|', '}', '~', '€'}

# from https://support.messagebird.com/hc/en-us/articles/208739745-How-long-is-1-SMS-Message-
MULTIPART_LENGTHS = [(1, 160), (2, 306), (3, 459), (4, 612), (5, 765), (6, 918), (7, 1071), (8, 1224), (9, 1377)]


@dataclass
class SmsLength:
    length: int
    parts: int


class MessageTooLong(ValueError):
    pass


def sms_length(msg: str) -> SmsLength:
    """
    :param msg: msg string
    :return: tuple (length of the message, number of multi-part SMSs required)
    """
    length = 0
    for c in msg:
        if c in BASIC_CHARACTERS:
            length += 1
        elif c in EXTENSION_CHARACTERS:
            length += 2
        # in theory all other characters are unavailable in GSM 03.38 and will be stripped out

    for msg_parts, max_length in MULTIPART_LENGTHS:
        if length <= max_length:
            return SmsLength(length, msg_parts)
    raise MessageTooLong(f'message length {length} exceeds maximum multi-part SMS length {max_length}')



import base64
import hashlib
import hmac
import json
import pytest
import re
from arq import Retry
from buildpg import V
from datetime import datetime, timedelta, timezone
from foxglove.db.helpers import SyncDb
from pathlib import Path
from pytest_toolbox.comparison import AnyInt, RegexStr
from starlette.testclient import TestClient
from uuid import uuid4

from src.schemas.messages import EmailRecipientModel
from src.worker import delete_old_emails, email_retrying, send_email as worker_send_email

THIS_DIR = Path(__file__).parent.resolve()


def test_send_email(cli: TestClient, worker, tmpdir, loop):
    uuid = str(uuid4())
    data = {
        'uid': uuid,
        'company_code': 'foobar',
        'from_address': 'Samuel <s@muelcolvin.com>',
        'method': 'email-test',
        'subject_template': 'test email {{ a }}',
        'context': {'message__render': '# hello\n\nThis is a **{{ b }}**.\n', 'a': 'Apple', 'b': 'Banana'},
        'recipients': [
            {
                'first_name': 'foo',
                'last_name': 'bar',
                'user_link': '/user/profile/42/',
                'address': 'foobar@example.org',
                'tags': ['foobar'],
            }
        ],
    }
    r = cli.post('/send/email/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(uuid + '-foobarexampleorg.txt').read()
    assert '\nsubject: test email Apple\n' in msg_file
    assert '\n<p>This is a <strong>Banana</strong>.</p>\n' in msg_file
    data = json.loads(re.search(r'data: ({.*?})\ncontent:', msg_file, re.S).groups()[0])
    assert data['from_email'] == 's@muelcolvin.com'
    assert data['to_address'] == 'foobar@example.org'
    assert data['to_user_link'] == '/user/profile/42/'
    assert data['attachments'] == []
    assert set(data['tags']) == {uuid, 'foobar'}


def test_webhook(cli: TestClient, send_email, sync_db: SyncDb, worker, loop):
    uuid = str(uuid4())
    message_id = send_email(uid=uuid)

    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == message_id)
    assert message['status'] == 'send'
    first_update_ts = message['update_ts']

    events = sync_db.fetchval('select count(*) from events')
    assert events == 0

    data = {'ts': int(2e9), 'event': 'open', '_id': message_id, 'foobar': ['hello', 'world']}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    assert worker.test_run() == 2

    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == message_id)
    assert message['status'] == 'open'
    assert message['update_ts'] > first_update_ts
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 1
    event = sync_db.fetchrow_b('select * from events where :where', where=V('message_id') == message['id'])
    assert event['ts'] == datetime(2033, 5, 18, 3, 33, 20, tzinfo=timezone.utc)
    assert event['extra'] == RegexStr('{.*}')
    extra = json.loads(event['extra'])
    assert extra['diag'] is None
    assert extra['opens'] is None


def test_webhook_old(cli: TestClient, send_email, sync_db: SyncDb, worker, loop):
    msg_id = send_email()
    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == msg_id)
    assert message['status'] == 'send'
    first_update_ts = message['update_ts']
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 0
    data = {'ts': int(1.4e9), 'event': 'open', '_id': msg_id}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    assert worker.test_run() == 2

    assert message['status'] == 'send'
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 1
    assert message['update_ts'] == first_update_ts


def test_webhook_repeat(cli: TestClient, send_email, sync_db: SyncDb, worker, loop):
    msg_id = send_email()
    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == msg_id)
    assert message['status'] == 'send'
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 0
    data = {'ts': '2032-06-06T12:10', 'event': 'open', '_id': msg_id}
    for _ in range(3):
        r = cli.post('/webhook/test/', json=data)
        assert r.status_code == 200, r.text
    data = {'ts': '2032-06-06T12:10', 'event': 'open', '_id': msg_id, 'user_agent': 'xx'}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    assert worker.test_run() == 5

    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == msg_id)
    assert message['status'] == 'open'
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 2


def test_webhook_missing(cli: TestClient, send_email, sync_db: SyncDb):
    msg_id = send_email()

    data = {'ts': int(1e10), 'event': 'open', '_id': 'missing', 'foobar': ['hello', 'world']}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    message = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == msg_id)
    assert message['status'] == 'send'
    assert sync_db.fetchval_b('select count(*) from events where :where', where=V('message_id') == message['id']) == 0


def test_mandrill_send(send_email, sync_db: SyncDb, dummy_server):
    assert sync_db.fetchval('select count(*) from messages') == 0
    send_email(method='email-mandrill', recipients=[{'address': 'foobar_a@testing.com'}])

    m = sync_db.fetchrow_b(
        'select * from messages where :where', where=V('external_id') == 'mandrill-foobaratestingcom'
    )
    assert m['to_address'] == 'foobar_a@testing.com'
    assert dummy_server.app['log'] == ['POST /mandrill/messages/send.json > 200']


def test_send_mandrill_with_other_attachments(send_email, sync_db: SyncDb, dummy_server):
    with open(THIS_DIR / 'attachments/testing.pdf', 'rb') as f:
        content = f.read()
    sent_content = base64.b64encode(content).decode()
    assert sync_db.fetchval('select count(*) from messages') == 0
    send_email(
        method='email-mandrill',
        recipients=[
            {
                'address': 'foobar_c@testing.com',
                'attachments': [
                    {'name': 'calendar.ics', 'content': 'Look this is some test data', 'mime_type': 'text/calendar'},
                    {'name': 'testing.pdf', 'content': sent_content, 'mime_type': 'application/pdf'},
                ],
            }
        ],
    )
    m = sync_db.fetchrow_b(
        'select * from messages where :where', where=V('external_id') == 'mandrill-foobarctestingcom'
    )
    assert m['to_address'] == 'foobar_c@testing.com'
    assert set(m['attachments']) == {'::calendar.ics', '::testing.pdf'}


def test_example_email_address(send_email, sync_db: SyncDb, dummy_server):
    assert sync_db.fetchval('select count(*) from messages') == 0
    send_email(method='email-mandrill', recipients=[{'address': 'foobar_a@example.com'}])

    m = sync_db.fetchrow_b(
        'select * from messages where :where', where=V('external_id') == 'mandrill-foobaraexamplecom'
    )
    assert m['to_address'] == 'foobar_a@example.com'
    assert m['status'] == 'send'


def test_mandrill_webhook(cli: TestClient, send_email, sync_db: SyncDb, worker, loop, dummy_server, settings):
    send_email(method='email-mandrill', recipients=[{'address': 'testing@example.org'}])
    assert sync_db.fetchval('select count(*) from messages') == 1

    assert sync_db.fetchval('select count(*) from events') == 0

    messages = [{'ts': 1969660800, 'event': 'open', '_id': 'mandrill-testingexampleorg', 'foobar': ['hello', 'world']}]

    msg = f'https://localhost/webhook/mandrill/mandrill_events{json.dumps(messages)}'
    sig = base64.b64encode(
        hmac.new(settings.mandrill_webhook_key.encode(), msg=msg.encode(), digestmod=hashlib.sha1).digest()
    )
    r = cli.post(
        '/webhook/mandrill/',
        data={'mandrill_events': json.dumps(messages)},
        headers={'X-Mandrill-Signature': sig.decode()},
    )
    assert r.status_code == 200, r.json()
    assert worker.test_run() == 2

    assert sync_db.fetchval('select count(*) from events') == 1

    events = sync_db.fetch('select * from events')
    assert events[0]['ts'] == datetime(2032, 6, 1, 0, 0, tzinfo=timezone.utc)
    assert events[0]['status'] == 'open'


def test_mandrill_webhook_invalid(cli: TestClient, send_email, sync_db: SyncDb, dummy_server, settings):
    send_email(method='email-mandrill', recipients=[{'address': 'testing@example.org'}])
    messages = [{'ts': 1969660800, 'event': 'open', '_id': 'e587306</div></body><meta name=', 'foobar': ['x']}]
    data = {'mandrill_events': messages}

    msg = (
        'https://localhost/webhook/mandrill/mandrill_events[{"ts": 1969660800, "event": "open", '
        '"_id": "e587306</div></body><meta name=", "foobar": ["x"]}]'
    )
    sig = base64.b64encode(
        hmac.new(settings.mandrill_webhook_key.encode(), msg=msg.encode(), digestmod=hashlib.sha1).digest()
    )
    r = cli.post('/webhook/mandrill/', data=data, headers={'X-Mandrill-Signature': sig.decode()})
    assert r.status_code == 400, r.text

    assert sync_db.fetchval('select count(*) from events') == 0


def test_mandrill_send_bad_template(cli: TestClient, send_email, sync_db: SyncDb, dummy_server):
    assert sync_db.fetchval('select count(*) from messages') == 0
    send_email(
        method='email-mandrill', main_template='{{ foo } test message', recipients=[{'address': 'foobar_b@testing.com'}]
    )
    message = sync_db.fetchrow_b('select * from messages')
    assert message['status'] == 'render_failed'


def test_send_email_headers(cli: TestClient, tmpdir, worker, loop, dummy_server):
    uid = str(uuid4())
    data = {
        'uid': uid,
        'company_code': 'foobar',
        'from_address': 'Samuel <s@muelcolvin.com>',
        'method': 'email-test',
        'subject_template': 'test email {{ a }}',
        'context': {'message__render': 'test email {{ a }} {{ b}} {{ c }}.\n', 'a': 'Apple', 'b': 'Banana'},
        'headers': {'Reply-To': 'another@whoever.com', 'List-Unsubscribe': '<http://example.org/unsub>'},
        'recipients': [
            {'first_name': 'foo', 'last_name': 'bar', 'address': 'foobar@example.org', 'context': {'c': 'Carrot'}},
            {
                'address': '2@example.org',
                'context': {'b': 'Banker'},
                'headers': {'List-Unsubscribe': '<http://example.org/different>'},
            },
        ],
    }
    r = cli.post('/send/email/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 2

    assert len(tmpdir.listdir()) == 2
    msg_file = tmpdir.join(f'{uid}-foobarexampleorg.txt').read()
    assert '<p>test email Apple Banana Carrot.</p>\n' in msg_file
    assert '"to_address": "foobar@example.org",\n' in msg_file
    assert '"Reply-To": "another@whoever.com",\n' in msg_file
    assert '"List-Unsubscribe": "<http://example.org/unsub>"\n' in msg_file

    msg_file = tmpdir.join(f'{uid}-2exampleorg.txt').read()
    assert '<p>test email Apple Banker .</p>\n' in msg_file
    assert '"to_address": "2@example.org",\n' in msg_file
    assert '"Reply-To": "another@whoever.com",\n' in msg_file
    assert '"List-Unsubscribe": "<http://example.org/different>"\n' in msg_file


def test_send_unsub_context(send_email, tmpdir):
    uid = str(uuid4())
    send_email(
        uid=uid,
        context={
            'message__render': 'test email {{ unsubscribe_link }}.\n',
            'unsubscribe_link': 'http://example.org/unsub',
        },
        recipients=[
            {'address': '1@example.org'},
            {
                'address': '2@example.org',
                'context': {'unsubscribe_link': 'http://example.org/context'},
                'headers': {'List-Unsubscribe': '<http://example.org/different>'},
            },
        ],
    )
    assert len(tmpdir.listdir()) == 2
    msg_file = tmpdir.join(f'{uid}-1exampleorg.txt').read()
    assert '"to_address": "1@example.org",\n' in msg_file
    assert '"List-Unsubscribe": "<http://example.org/unsub>"\n' in msg_file
    assert '<p>test email http://example.org/unsub.</p>\n' in msg_file

    msg_file = tmpdir.join(f'{uid}-2exampleorg.txt').read()
    assert '"to_address": "2@example.org",\n' in msg_file
    assert '"List-Unsubscribe": "<http://example.org/different>"\n' in msg_file
    assert '<p>test email http://example.org/context.</p>\n' in msg_file


def test_markdown_context(send_email, tmpdir):
    message_id = send_email(
        main_template='testing {{{ foobar }}}',
        context={
            'message__render': 'test email {{ unsubscribe_link }}.\n',
            'foobar__md': '[hello](www.example.org/hello)',
        },
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert 'content:\ntesting <p><a href="www.example.org/hello">hello</a></p>\n' in msg_file


def test_partials(send_email, tmpdir):
    message_id = send_email(
        main_template='message: |{{{ message }}}|\n' 'foo: {{ foo }}\n' 'partial: {{> test_p }}',
        context={'message__render': '{{foo}} {{> test_p }}', 'foo': 'FOO', 'bar': 'BAR'},
        mustache_partials={'test_p': 'foo ({{ foo }}) bar **{{ bar }}**'},
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert (
        """
content:
message: |<p>FOO foo (FOO) bar <strong>BAR</strong></p>
|
foo: FOO
partial: foo (FOO) bar **BAR**
"""
        in msg_file
    )


def test_macros(send_email, tmpdir):
    message_id = send_email(
        main_template='macro result: foobar(hello | {{ foo }})',
        context={'foo': 'FOO', 'bar': 'BAR'},
        macros={'foobar(a | b)': '___{{ a }} {{b}}___'},
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert 'content:\nmacro result: ___hello FOO___\n' in msg_file


def test_macros_more(send_email, tmpdir):
    message_id = send_email(
        main_template=(
            'foo:foo()\n'
            'foo wrong:foo(1 | 2)\n'
            'bar:bar()\n'
            'spam1:spam(x | y )\n'
            'spam2:spam(with bracket )  | {{ bar}} )\n'
            'spam3:spam({{ foo }} | {{ bar}} )\n'
            'spam wrong:spam(1 | {{ bar}} | x)\n'
            'button:centered_button(Reset password now | {{ password_reset_link }})\n'
        ),
        context={
            'foo': 'FOO',
            'bar': 'BAR',
            'password_reset_link': '/testagency/password/reset/t-4mx-2968ca2f34bc512e70e6/',
        },
        macros={
            'foo()': '___is foo___',
            'bar': '___is bar___',
            'spam(apple | pear)': '___spam {{apple}} {{pear}}___',
            'centered_button(text | link)': """
      <div class="button">
        <a href="{{ link }}"><span>{{ text }}</span></a>
      </div>\n""",
        },
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert (
        """
content:
foo:___is foo___
foo wrong:foo(1 | 2)
bar:bar()
spam1:___spam x y___
spam2:spam(with bracket )  | BAR )
spam3:___spam FOO BAR___
spam wrong:spam(1 | BAR | x)
button:
      <div class="button">
        <a href="/testagency/password/reset/t-4mx-2968ca2f34bc512e70e6/"><span>Reset password now</span></a>
      </div>
"""
        in msg_file
    )


def test_macro_in_message(send_email, tmpdir):
    message_id = send_email(
        context={
            'pay_link': '/pay/now/123/',
            'first_name': 'John',
            'message__render': '# hello {{ first_name }}\n' 'centered_button(Pay now | {{ pay_link }})\n',
        },
        macros={
            'centered_button(text | link)': (
                '<div class="button">\n' '  <a href="{{ link }}"><span>{{ text }}</span></a>\n' '</div>\n'
            )
        },
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert (
        """
content:
<body>
<h1>hello John</h1>

<div class="button">
  <a href="/pay/now/123/"><span>Pay now</span></a>
</div>

</body>
"""
        in msg_file
    )


def test_send_md_options(send_email, tmpdir):
    message_id = send_email(context={'message__render': 'we are_testing_emphasis **bold**\nnewline'})
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert '<p>we are_testing_emphasis <strong>bold</strong><br>\nnewline</p>' in msg_file


def test_standard_sass(cli: TestClient, tmpdir, worker, loop):
    data = dict(
        uid=str(uuid4()),
        company_code='foobar',
        from_address='Sender Name <sender@example.org>',
        method='email-test',
        subject_template='test message',
        context={'message': 'this is a test'},
        recipients=[{'address': 'foobar@testing.com'}],
    )
    r = cli.post('/send/email/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201
    assert worker.test_run() == 1
    message_id = data['uid'] + '-foobartestingcom'

    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert '<style>#body{-webkit-font-smoothing' in msg_file


def test_custom_sass(send_email, tmpdir):
    message_id = send_email(
        main_template='{{{ css }}}',
        context={'css__sass': '.foo {\n  .bar {\n    color: black;\n    width: (60px / 6);\n  }\n' '}'},
    )

    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert '.foo .bar{color:black;width:10px}' in msg_file
    assert '#body{-webkit-font-smoothing' not in msg_file


def test_invalid_mustache_subject(send_email, tmpdir, sync_db: SyncDb):
    message_id = send_email(
        subject_template='{{ foo } test message', context={'foo': 'FOO'}, company_code='test_invalid_mustache_subject'
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert '\nsubject: {{ foo } test message\n' in msg_file

    message = sync_db.fetchrow_b('select * from messages')
    assert message['status'] == 'send'
    assert message['subject'] == '{{ foo } test message'
    assert message['body'] == '<body>\n\n</body>'


def test_invalid_mustache_body(send_email, sync_db: SyncDb):
    send_email(main_template='{{ foo } test message', context={'foo': 'FOO'}, company_code='test_invalid_mustache_body')

    m = sync_db.fetchrow_b('select * from messages')
    assert m['status'] == 'render_failed'
    assert m['subject'] is None
    assert m['body'] == 'Error rendering email: unclosed tag at line 1'


def test_send_with_pdf(send_email, tmpdir, sync_db: SyncDb):
    message_id = send_email(
        recipients=[
            {
                'address': 'foobar@testing.com',
                'pdf_attachments': [
                    {'name': 'testing.pdf', 'html': '<h1>testing</h1>', 'id': 123},
                    {'name': 'different.pdf', 'html': '<h1>different</h1>'},
                ],
            }
        ]
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert 'testing.pdf' in msg_file

    attachments = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == message_id)[
        'attachments'
    ]
    assert set(attachments) == {'123::testing.pdf', '::different.pdf'}


def test_send_with_other_attachment(send_email, tmpdir, sync_db: SyncDb):
    message_id = send_email(
        recipients=[
            {
                'address': 'foobar@testing.com',
                'attachments': [
                    {'name': 'calendar.ics', 'content': 'Look this is some test data', 'mime_type': 'text/calendar'}
                ],
            }
        ]
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert 'Look this is some test data' in msg_file
    attachments = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == message_id)[
        'attachments'
    ]
    assert set(attachments) == {'::calendar.ics'}


def test_send_with_other_attachment_pdf(send_email, tmpdir, sync_db: SyncDb):
    msg = 'Look this is some test data'
    encoded_content = base64.b64encode(b'Look this is some test data').decode()
    message_id = send_email(
        recipients=[
            {
                'address': 'foobar@testing.com',
                'attachments': [
                    {'name': 'test_pdf.pdf', 'content': msg, 'mime_type': 'application/pdf'},
                    {'name': 'test_pdf_encoded.pdf', 'content': encoded_content, 'mime_type': 'application/pdf'},
                ],
            }
        ]
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert f'test_pdf.pdf:{msg}' in msg_file
    assert f'test_pdf_encoded.pdf:{msg}' in msg_file
    attachments = sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == message_id)[
        'attachments'
    ]
    assert set(attachments) == {'::test_pdf.pdf', '::test_pdf_encoded.pdf'}


def test_pdf_not_unicode(send_email, tmpdir, cli):
    message_id = send_email(
        recipients=[
            {'address': 'foobar@testing.com', 'pdf_attachments': [{'name': 'testing.pdf', 'html': '<h1>binary</h1>'}]}
        ]
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert 'testing.pdf' in msg_file


def test_pdf_empty(send_email, tmpdir, dummy_server):
    message_id = send_email(
        recipients=[{'address': 'foobar@testing.com', 'pdf_attachments': [{'name': 'testing.pdf', 'html': ''}]}]
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{message_id}.txt').read()
    assert '\n  "attachments": []\n' in msg_file


def test_mandrill_send_client_error(sync_db: SyncDb, worker_ctx, call_send_emails, loop):
    group_id, c_id, m = call_send_emails(subject_template='__slow__')

    assert sync_db.fetchval('select count(*) from messages') == 0
    worker_ctx['job_try'] = 1

    with pytest.raises(Retry) as exc_info:
        loop.run_until_complete(
            worker_send_email(worker_ctx, group_id, c_id, EmailRecipientModel(address='testing@recipient.com'), m)
        )
    assert exc_info.value.defer_score == 5_000

    assert sync_db.fetchval('select count(*) from messages') == 0


def test_mandrill_send_many_errors(sync_db: SyncDb, worker_ctx, call_send_emails, loop):
    group_id, c_id, m = call_send_emails()

    assert sync_db.fetchval('select count(*) from messages') == 0
    worker_ctx['job_try'] = 10

    loop.run_until_complete(
        worker_send_email(worker_ctx, group_id, c_id, EmailRecipientModel(address='testing@recipient.com'), m)
    )

    m = sync_db.fetchrow_b('select * from messages')
    assert m['status'] == 'send_request_failed'
    assert m['body'] == 'upstream error'


def test_mandrill_send_502(sync_db: SyncDb, call_send_emails, loop, worker_ctx):
    group_id, c_id, m = call_send_emails(subject_template='__502__')

    worker_ctx['job_try'] = 1

    with pytest.raises(Retry) as exc_info:
        loop.run_until_complete(
            worker_send_email(worker_ctx, group_id, c_id, EmailRecipientModel(address='testing@recipient.com'), m)
        )
    assert exc_info.value.defer_score == 5_000

    assert sync_db.fetchval('select count(*) from messages') == 0


def test_mandrill_send_502_last(sync_db: SyncDb, call_send_emails, loop, worker_ctx):
    group_id, c_id, m = call_send_emails(subject_template='__502__')

    worker_ctx['job_try'] = len(email_retrying)

    with pytest.raises(Retry) as exc_info:
        loop.run_until_complete(
            worker_send_email(worker_ctx, group_id, c_id, EmailRecipientModel(address='testing@recipient.com'), m)
        )
    assert exc_info.value.defer_score == 43_200_000

    assert sync_db.fetchval('select count(*) from messages') == 0


def test_mandrill_send_500_nginx(sync_db: SyncDb, call_send_emails, loop, worker_ctx):
    group_id, c_id, m = call_send_emails(subject_template='__500_nginx__')

    worker_ctx['job_try'] = 2

    with pytest.raises(Retry) as exc_info:
        loop.run_until_complete(
            worker_send_email(worker_ctx, group_id, c_id, EmailRecipientModel(address='testing@recipient.com'), m)
        )
    assert exc_info.value.defer_score == 10_000
    assert sync_db.fetchval('select count(*) from messages') == 0


def send_with_link(send_email, tmpdir):
    mid = send_email(
        main_template='<a href="{{ the_link }}">foobar</a> test message',
        context={'the_link': 'https://www.foobar.com'},
        company_code='test_link_shortening',
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{mid}.txt').read()
    m = re.search(r'<a href="https://click.example.com/l(.+?)\?u=(.+?)">foobar</a> test message', msg_file)
    assert m, msg_file
    token, enc_url = m.groups()
    assert len(token) == 30
    assert base64.urlsafe_b64decode(enc_url).decode() == 'https://www.foobar.com'
    return token


def test_link_shortening(send_email, tmpdir, cli: TestClient, sync_db: SyncDb, worker, loop):
    token = send_with_link(send_email, tmpdir)

    m = sync_db.fetchrow_b('select * from messages')
    assert m['status'] == 'send'

    link = sync_db.fetchrow_b('select * from links')
    assert link['id'] == AnyInt()
    assert link['message_id'] == m['id']
    assert link['token'] == token
    assert link['url'] == 'https://www.foobar.com'

    r = cli.get(
        '/l' + token,
        allow_redirects=False,
        headers={
            'X-Forwarded-For': '54.170.228.0, 141.101.88.55',
            'X-Request-Start': '1969660800',
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
            'Chrome/59.0.3071.115 Safari/537.36',
        },
    )
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'https://www.foobar.com'
    assert worker.test_run() == 2

    m = sync_db.fetchrow_b('select * from messages where :where', where=V('id') == m['id'])
    assert m['status'] == 'click'
    event = sync_db.fetchrow_b('select * from events')
    assert event['status'] == 'click'
    assert event['ts'] == datetime(2032, 6, 1, 0, 0, tzinfo=timezone.utc)
    extra = json.loads(event['extra'])
    assert extra == {
        'ip': '54.170.228.0',
        'target': 'https://www.foobar.com',
        'user_agent': (
            'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
            'Chrome/59.0.3071.115 Safari/537.36'
        ),
        'user_agent_display': 'Chrome 59 on Linux',
    }


def test_link_shortening_wrong_url(send_email, tmpdir, cli, dummy_server):
    token = send_with_link(send_email, tmpdir)
    # check we use the right url with a valid token but a different url arg
    r = cli.get('/l' + token + '?u=' + base64.urlsafe_b64encode(b'different').decode(), allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'https://www.foobar.com'


def test_link_shortening_wrong_url_missing(send_email, tmpdir, cli, dummy_server):
    token = send_with_link(send_email, tmpdir)
    r = cli.get('/lx' + token + '?u=' + base64.urlsafe_b64encode(b'different').decode(), allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'different'


def test_link_shortening_repeat(send_email, tmpdir, cli: TestClient, sync_db: SyncDb, worker, loop, dummy_server):
    token = send_with_link(send_email, tmpdir)
    r = cli.get('/l' + token, allow_redirects=False)
    assert r.status_code == 307, r.text
    assert worker.test_run() == 2
    assert r.headers['location'] == 'https://www.foobar.com'
    assert sync_db.fetchval('select count(*) from events') == 1

    r = cli.get('/l' + token, allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'https://www.foobar.com'
    assert sync_db.fetchval('select count(*) from events') == 1


def test_link_shortening_in_render(send_email, tmpdir, sync_db: SyncDb):
    mid = send_email(
        context={'message__render': 'test email {{ xyz }}\n', 'xyz': 'http://example.com/foobar'},
        company_code='test_link_shortening_in_render',
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{mid}.txt').read()
    m = re.search(r'<p>test email https://click.example.com/l(.+?)\?u=(.+?)</p>', msg_file)
    assert m, msg_file
    token, enc_url = m.groups()

    link = sync_db.fetchrow_b('select * from links')
    assert link['url'] == 'http://example.com/foobar'
    assert link['token'] == token
    assert base64.urlsafe_b64decode(enc_url).decode() == 'http://example.com/foobar'


def test_link_shortening_keep_long_link(send_email, tmpdir, cli):
    mid = send_email(
        context={'message__render': 'test email {{ xyz_original }}\n', 'xyz': 'http://example.org/foobar'},
        company_code='test_link_shortening_in_render',
    )
    msg_file = tmpdir.join(f'{mid}.txt').read()
    m = re.search(r'<p>test email http://example.org/foobar</p>', msg_file)
    assert m, msg_file


def test_link_shortening_not_image(send_email, tmpdir, cli):
    mid = send_email(
        context={
            'message__render': '{{ foo }} {{ bar}}',
            'foo': 'http://example.com/foobar',
            'bar': 'http://whatever.com/img.jpg',
        },
        company_code='test_link_shortening_in_render',
    )
    assert len(tmpdir.listdir()) == 1
    msg_file = tmpdir.join(f'{mid}.txt').read()
    assert re.search(r'<p>https://click.example.com/l(\S+) http://whatever\.com/img\.jpg</p>', msg_file), msg_file


def test_not_json(cli: TestClient, tmpdir):
    r = cli.post('/send/email/', data='xxx', headers={'Authorization': 'testing-key'})
    assert r.status_code == 422, r.text


def test_invalid_json(cli: TestClient, tmpdir):
    data = {
        'uid': 'xxx',
        'company_code': 'foobar',
        'from_address': 'Samuel <s@muelcolvin.com>',
        'method': 'email-test',
        'subject_template': 'test email {{ a }}',
        'context': {},
        'recipients': [{'address': 'foobar_a@testing.com'}],
    }
    r = cli.post('/send/email/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 422
    assert {
        'detail': [{'loc': ['body', 'uid'], 'msg': 'value is not a valid uuid', 'type': 'type_error.uuid'}]
    } == r.json()


def test_delete_old_messages(cli: TestClient, send_email, sync_db: SyncDb, worker, loop):
    for i in range(3):
        send_email()

    m = sync_db.fetch('select * from messages')[0]
    sync_db.execute(
        'update message_groups set created_ts = $1 where id = $2', datetime.today() - timedelta(days=366), m['group_id']
    )
    m = sync_db.fetch('select * from messages')[1]
    sync_db.execute(
        'update message_groups set created_ts = $1 where id = $2', datetime.today() - timedelta(days=200), m['group_id']
    )
    m = sync_db.fetch('select * from messages')[2]
    sync_db.execute(
        'update message_groups set created_ts = $1 where id = $2', datetime.today() + timedelta(days=1), m['group_id']
    )

    assert sync_db.fetchval('select count(*) from messages') == 3
    loop.run_until_complete(delete_old_emails({'pg': sync_db}))
    assert sync_db.fetchval('select count(*) from messages') == 2



import logging
import re
from datetime import datetime, timedelta
from foxglove.db.helpers import SyncDb
from urllib.parse import urlencode
from uuid import uuid4


def test_send_message(cli, tmpdir, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d71',
        'company_code': 'foobar',
        'method': 'sms-test',
        'from_name': 'foobar send',
        'main_template': 'this is a message {{ foo }}',
        'recipients': [{'number': '07891123856', 'context': {'foo': 'bar'}}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    f = '69eb85e8-1504-40aa-94ff-75bb65fd8d71-447891123856.txt'
    assert str(tmpdir.listdir()[0]).endswith(f)
    msg_file = tmpdir.join(f).read()
    assert (
        "to: Number(number='+447891123856', country_code='44', "
        "number_formatted='+44 7891 123856', descr=None, is_mobile=True)"
    ) in msg_file
    assert '\nfrom_name: foobar send\n' in msg_file
    assert '\nmessage:\nthis is a message bar\n' in msg_file
    assert '\nlength: SmsLength(length=21, parts=1)\n' in msg_file


def test_send_message_usa(cli, settings, tmpdir, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d72',
        'company_code': 'foobar',
        'country_code': 'US',
        'from_name': 'foobar send',
        'method': 'sms-test',
        'main_template': 'this is a message {{ foo }}',
        'recipients': [{'number': '+1 818 337 3095', 'context': {'foo': 'bar'}}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    f = '69eb85e8-1504-40aa-94ff-75bb65fd8d72-18183373095.txt'
    assert str(tmpdir.listdir()[0]).endswith(f)
    msg_file = tmpdir.join(f).read()
    assert (
        "to: Number(number='+18183373095', country_code='1', "
        "number_formatted='+1 818-337-3095', descr=None, is_mobile=True)"
    ) in msg_file
    assert f'\nfrom_name: {settings.us_send_number}\n' in msg_file
    assert '\nmessage:\nthis is a message bar\n' in msg_file
    assert '\nlength: SmsLength(length=21, parts=1)\n' in msg_file


def test_validate_number(cli, tmpdir):
    data = {
        'country_code': 'US',
        'numbers': {
            123: 'xxxxx',
            234: '1 818 337 3095',
            345: '+447891123856',
            456: '+44 (0) 207 1128 953',
            567: '+12001230101',  # not possible
        },
    }
    r = cli.get('/validate/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    data = r.json()
    assert {
        '123': None,
        '234': {
            'number': '+18183373095',
            'country_code': '1',
            'number_formatted': '+1 818-337-3095',
            'descr': 'California, United States',
            'is_mobile': True,
        },
        '345': {
            'number': '+447891123856',
            'country_code': '44',
            'number_formatted': '+44 7891 123856',
            'descr': 'United Kingdom',
            'is_mobile': True,
        },
        '456': {
            'number': '+442071128953',
            'country_code': '44',
            'number_formatted': '+44 20 7112 8953',
            'descr': 'London, United Kingdom',
            'is_mobile': False,
        },
        '567': None,
    } == data


def test_repeat_uuid(cli, tmpdir, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d73',
        'company_code': 'foobar',
        'method': 'sms-test',
        'main_template': 'this is a message',
        'recipients': [{'number': '07891123856'}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    assert str(tmpdir.listdir()[0]).endswith('69eb85e8-1504-40aa-94ff-75bb65fd8d73-447891123856.txt')
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 409, r.text
    data = r.json()
    assert {'message': 'Send group with id "69eb85e8-1504-40aa-94ff-75bb65fd8d73" already exists\n'} == data


def test_invalid_number(cli, tmpdir, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d74',
        'company_code': 'foobar',
        'country_code': 'US',
        'method': 'sms-test',
        'main_template': 'this is a message',
        'recipients': [
            {'number': '+447891123856'},  # uk mobile
            {'number': '+44 (0) 207 1128 953'},  # not mobile
            {'number': '1 818 337 3095'},  # US mobile or fix
            {'number': '+12001230101'},  # not possible
        ],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 4
    assert len(tmpdir.listdir()) == 2
    files = {str(f).split('/')[-1] for f in tmpdir.listdir()}
    assert files == {
        '69eb85e8-1504-40aa-94ff-75bb65fd8d74-18183373095.txt',
        '69eb85e8-1504-40aa-94ff-75bb65fd8d74-447891123856.txt',
    }


def test_exceed_cost_limit(cli, tmpdir, worker, loop):
    d = {
        'company_code': 'cost-test',
        'cost_limit': 0.1,
        'method': 'sms-test',
        'main_template': 'this is a message',
        'recipients': [{'number': f'0789112385{i}'} for i in range(4)],
    }
    r = cli.post('/send/sms/', json=dict(uid=str(uuid4()), **d), headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 4
    assert {'status': 'enqueued', 'spend': 0.0} == r.json()
    assert len(tmpdir.listdir()) == 4
    r = cli.post('/send/sms/', json=dict(uid=str(uuid4()), **d), headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 8
    assert {'status': 'enqueued', 'spend': 0.048} == r.json()
    assert len(tmpdir.listdir()) == 8

    r = cli.post('/send/sms/', json=dict(uid=str(uuid4()), **d), headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    obj = r.json()
    assert 0.095 < obj['spend'] < 0.097
    assert worker.test_run() == 12
    assert len(tmpdir.listdir()) == 12

    r = cli.post('/send/sms/', json=dict(uid=str(uuid4()), **d), headers={'Authorization': 'testing-key'})
    assert r.status_code == 402, r.text
    obj = r.json()
    assert 0.143 < obj['spend'] < 0.145
    assert obj['cost_limit'] == 0.1
    assert len(tmpdir.listdir()) == 12


def test_send_messagebird(cli, tmpdir, dummy_server, worker, loop):
    data = {
        'uid': str(uuid4()),
        'company_code': 'foobar',
        'method': 'sms-messagebird',
        'main_template': 'this is a message',
        'recipients': [{'number': '07801234567'}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert [
        'POST /messagebird/lookup/447801234567/hlr > 201',
        'GET /messagebird/lookup/447801234567 > 200',
        'GET /messagebird/pricing/sms/outbound > 200',
        'POST /messagebird/messages > 201',
    ] == dummy_server.log

    # send again, this time hlr look and pricing requests shouldn't occur
    data = dict(data, uid=str(uuid4()))
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 2
    assert len(dummy_server.log) == 5
    assert dummy_server.log[4] == 'POST /messagebird/messages > 201'


def test_messagebird_no_hlr(cli, tmpdir, dummy_server, worker, caplog, loop):
    data = {
        'uid': str(uuid4()),
        'company_code': 'foobar',
        'method': 'sms-messagebird',
        'main_template': 'this is a message',
        'recipients': [{'number': '07888888888'}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert [
        'POST /messagebird/lookup/447888888888/hlr > 201',
        *['GET /messagebird/lookup/447888888888 > 200' for _ in range(30)],
    ] == dummy_server.log
    dummy_server.log = []
    assert 'No HLR result found for +447888888888 after 30 attempts' in caplog.messages


def test_messagebird_no_network(cli, tmpdir, dummy_server, worker, caplog, loop):
    data = {
        'uid': str(uuid4()),
        'company_code': 'foobar',
        'method': 'sms-messagebird',
        'main_template': 'this is a message',
        'recipients': [{'number': '07777777777'}],
    }
    caplog.set_level(logging.INFO)
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert [
        'POST /messagebird/lookup/447777777777/hlr > 201',
        'GET /messagebird/lookup/447777777777 > 200',
        'GET /messagebird/lookup/447777777777 > 200',
        'GET /messagebird/pricing/sms/outbound > 200',
        'POST /messagebird/messages > 201',
    ] == dummy_server.log
    dummy_server.log = []
    assert (
        """found result for +447777777777 after 1 attempts {
  "hlr": {
    "status": "active",
    "network": "o2"
  }
}"""
        in caplog.messages
    )


def test_messagebird_webhook(cli, sync_db: SyncDb, dummy_server, worker, loop):
    data = {
        'uid': str(uuid4()),
        'company_code': 'webhook-test',
        'method': 'sms-messagebird',
        'main_template': 'this is a message',
        'recipients': [{'first_name': 'John', 'last_name': 'Doe', 'user_link': 4321, 'number': '07801234567'}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1

    msg = sync_db.fetchrow_b('select * from messages join message_groups g on g.id = messages.id')
    assert msg['status'] == 'send'
    assert msg['to_first_name'] == 'John'
    assert msg['to_last_name'] == 'Doe'
    assert msg['to_user_link'] == '4321'
    assert msg['to_address'] == '+44 7801 234567'
    assert msg['from_name'] == 'Morpheus'
    assert msg['body'] == 'this is a message'
    assert msg['cost'] == 0.02
    assert len(msg['tags']) == 1  # just group_id

    url_args = {
        'id': msg['external_id'],
        'reference': 'morpheus',
        'recipient': '447801234567',
        'status': 'delivered',
        'statusDatetime': '2032-06-06T12:00:00',
    }
    r = cli.get(f'/webhook/messagebird/?{urlencode(url_args)}')
    assert r.status_code == 200, r.text
    assert worker.test_run() == 2

    msg = sync_db.fetchrow_b('select * from messages')
    assert msg['status'] == 'delivered'


def test_failed_render(cli, tmpdir, sync_db: SyncDb, worker, loop):
    data = {
        'uid': str(uuid4()),
        'company_code': 'test_failed_render',
        'method': 'sms-test',
        'context': {'foo': 'FOO'},
        'main_template': 'this is a message {{ foo }',
        'recipients': [{'number': '07891123856'}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 0

    assert sync_db.fetchrow_b('select * from messages')['status'] == 'render_failed'


def test_link_shortening(cli, tmpdir, sync_db: SyncDb, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d75',
        'company_code': 'sms_test_link_shortening',
        'method': 'sms-test',
        'main_template': 'this is a message {{ foo }}',
        'recipients': [{'number': '07891123856', 'context': {'foo': 'http://whatever.com/foo/bar'}}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    f = '69eb85e8-1504-40aa-94ff-75bb65fd8d75-447891123856.txt'
    assert str(tmpdir.listdir()[0]).endswith(f)
    msg_file = tmpdir.join(f).read()
    assert '\nfrom_name: Morpheus\n' in msg_file
    assert '\nmessage:\nthis is a message click.example.com/l' in msg_file
    token = re.search('message click.example.com/l(.+?)\n', msg_file).groups()[0]
    assert len(token) == 12

    link = sync_db.fetchrow_b('select * from links')
    assert link['url'] == 'http://whatever.com/foo/bar'
    assert link['token'] == token

    r = cli.get(f'/l{token}', allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'http://whatever.com/foo/bar'

    r = cli.get(f'/l{token}.', allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['location'] == 'http://whatever.com/foo/bar'


def test_send_multi_part(cli, tmpdir, worker, loop):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d76',
        'company_code': 'foobar',
        'method': 'sms-test',
        'main_template': 'this is a message {{ foo }}\n' * 10,
        'recipients': [{'number': '07891123856', 'context': {'foo': 'bar'}}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert worker.test_run() == 1
    assert len(tmpdir.listdir()) == 1
    f = '69eb85e8-1504-40aa-94ff-75bb65fd8d76-447891123856.txt'
    assert str(tmpdir.listdir()[0]).endswith(f)
    msg_file: str = tmpdir.join(f).read()
    assert '\nlength: SmsLength(length=230, parts=2)\n' in msg_file
    assert msg_file.count('this is a message bar') == 10


def test_send_too_long(cli, tmpdir):
    data = {
        'uid': '69eb85e8-1504-40aa-94ff-75bb65fd8d77',
        'company_code': 'foobar',
        'method': 'sms-test',
        'main_template': 'x' * 1500,
        'recipients': [{'number': '07891123856', 'context': {'foo': 'bar'}}],
    }
    r = cli.post('/send/sms/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    # no messages sent:
    assert len(tmpdir.listdir()) == 0


def test_sms_billing(cli, send_sms):
    for i in range(4):
        send_sms(uid=str(uuid4()), company_code='billing-test')

    start = (datetime.utcnow() - timedelta(days=5)).strftime('%Y-%m-%d')
    end = (datetime.utcnow() + timedelta(days=5)).strftime('%Y-%m-%d')
    data = dict(start=start, end=end, company_code='billing-test')
    r = cli.get(
        '/billing/sms-test/billing-test/', json=dict(uid=str(uuid4()), **data), headers={'Authorization': 'testing-key'}
    )
    assert r.status_code == 200, r.text
    assert {'company': 'billing-test', 'start': start, 'end': end, 'spend': 0.048} == r.json()



import pytest

from src.render.main import MessageTooLong, SmsLength, sms_length


def idfn(v):
    if isinstance(v, str):
        return f'{v[:20]}:{len(v)}'


@pytest.mark.parametrize(
    'msg, length, sms_count',
    [
        ('123', 3, 1),
        ('123\n456', 8, 1),
        ('123😀', 3, 1),
        ('123®', 4, 1),
        ('123{', 5, 1),
        ('{}', 4, 1),
        ('a' * 160, 160, 1),
        ('b' * 161, 161, 2),
        ('c' * 306, 306, 2),
        ('d' * 307, 307, 3),
        ('e' * 1377, 1377, 9),
        ('{' * 100, 200, 2),
    ],
    ids=idfn,
)
def test_sms_lengths(msg, length, sms_count):
    assert sms_length(msg) == SmsLength(length, sms_count)


def test_sms_too_long():
    with pytest.raises(MessageTooLong) as exc_info:
        sms_length('x' * 1378)
    assert exc_info.value.args[0] == 'message length 1378 exceeds maximum multi-part SMS length 1377'



import hashlib
import hmac
import json
import uuid
from buildpg import V, Values
from datetime import date, datetime, timedelta, timezone
from foxglove import glove
from foxglove.db.helpers import SyncDb
from operator import itemgetter
from pytest_toolbox.comparison import RegexStr
from starlette.testclient import TestClient
from urllib.parse import urlencode

from src.schemas.messages import MessageStatus


def modify_url(url, settings, company='foobar'):
    args = dict(company=company, expires=round(datetime(2032, 1, 1).timestamp()))
    body = '{company}:{expires}'.format(**args).encode()
    args['signature'] = hmac.new(settings.user_auth_key, body, hashlib.sha256).hexdigest()
    return str(url) + ('&' if '?' in str(url) else '?') + urlencode(args)


def test_user_list(cli, settings, send_email, sync_db: SyncDb):
    expected_msg_ids = []
    for i in range(4):
        uid = str(uuid.uuid4())
        send_email(uid=uid, company_code='whoever', recipients=[{'address': f'{i}@t.com'}])
        expected_msg_ids.append(f'{uid}-{i}tcom')

    send_email(uid=str(uuid.uuid4()), company_code='different1')
    send_email(uid=str(uuid.uuid4()), company_code='different2')
    r = cli.get(modify_url('/messages/email-test/', settings, 'whoever'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 4
    msg_ids = [h['external_id'] for h in data['items']]
    assert msg_ids == list(reversed(expected_msg_ids))
    first_item = data['items'][0]
    assert first_item == {
        'id': sync_db.fetchrow_b('select * from messages where :where', where=V('external_id') == expected_msg_ids[3])[
            'id'
        ],
        'external_id': expected_msg_ids[3],
        'to_ext_link': None,
        'to_address': '3@t.com',
        'to_dst': '<3@t.com>',
        'to_name': ' ',
        'send_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'update_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'status': 'Sent',
        'method': 'email-test',
        'subject': 'test message',
        'cost': 0,
    }


def test_user_list_no_ext(cli, settings, send_email, sync_db: SyncDb):
    send_email(
        uid=str(uuid.uuid4()),
        company_code='testing',
        recipients=[{'address': '3@t.com'}],
        subject_template='test message',
    )
    sync_db.execute_b('update messages set external_id=null')
    r = cli.get(modify_url('/messages/email-test/', settings, 'testing'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1
    first_item = data['items'][0]
    assert first_item == {
        'id': sync_db.fetchrow_b('select * from messages')['id'],
        'external_id': None,
        'to_ext_link': None,
        'to_address': '3@t.com',
        'to_dst': '<3@t.com>',
        'to_name': ' ',
        'send_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'update_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'status': 'Sent',
        'method': 'email-test',
        'subject': 'test message',
        'cost': 0,
    }


def test_user_search(cli, settings, send_email):
    msgs = {}
    for i, subject in enumerate(['apple', 'banana', 'cherry', 'durian']):
        uid = str(uuid.uuid4())
        send_email(uid=uid, company_code='whoever', recipients=[{'address': f'{i}@t.com'}], subject_template=subject)
        msgs[subject] = f'{uid}-{i}tcom'

    send_email(uid=str(uuid.uuid4()), company_code='different1', subject_template='eggplant')

    r = cli.get(modify_url('/messages/email-test/?q=cherry', settings, 'whoever'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1
    item = data['items'][0]
    assert item['external_id'] == msgs['cherry']
    assert item['subject'] == 'cherry'
    r = cli.get(modify_url('/messages/email-test/?q=eggplant', settings, 'whoever'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 0


def test_user_search_space(cli, settings, send_email):
    send_email(
        company_code='testing',
        recipients=[{'address': 'testing@example.com'}],
        subject_template='foobar',
    )
    send_email(
        company_code='testing',
        recipients=[{'address': 'testing@example.com'}],
        subject_template='bar',
    )
    send_email(
        company_code='testing',
        recipients=[{'address': 'testing@example.com'}],
        subject_template='foo bar',
    )

    r = cli.get(modify_url('/messages/email-test/?q=foobar', settings, 'testing'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1

    r = cli.get(modify_url('/messages/email-test/?q=foo%20bar', settings, 'testing'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1


def test_pagination(cli, settings, send_email):
    for i in range(110):
        send_email(
            uid=str(uuid.uuid4()),
            company_code='testing',
            recipients=[{'address': f'{i}@t.com'}],
            subject_template='foobar',
        )

    for i in range(20):
        send_email(
            uid=str(uuid.uuid4()),
            company_code='testing',
            recipients=[{'address': f'{i}@t.com'}],
            subject_template='barfoo',
        )

    r = cli.get(modify_url('/messages/email-test/', settings, 'testing'))
    assert r.status_code == 200, r.text
    data = r.json()
    first_item = data['items'][0]
    assert len(data['items']) == 100
    assert data['count'] == 130

    r = cli.get(modify_url('/messages/email-test/', settings, 'testing') + '&offset=100')
    assert r.status_code == 200, r.text
    data = r.json()
    assert first_item not in data['items']
    assert len(data['items']) == 30
    assert data['count'] == 130


def test_user_aggregate(cli, settings, send_email, sync_db: SyncDb, loop, worker):
    for i in range(4):
        send_email(uid=str(uuid.uuid4()), company_code='user-aggs', recipients=[{'address': f'{i}@t.com'}])
    msg_id = send_email(uid=str(uuid.uuid4()), company_code='user-aggs', recipients=[{'address': f'{i}@t.com'}])

    data = {'ts': int(2e10), 'event': 'open', '_id': msg_id, 'user_agent': 'testincalls'}
    cli.post('/webhook/test/', json=data)

    send_email(uid=str(uuid.uuid4()), company_code='different')
    loop.run_until_complete(glove.redis.enqueue_job('update_aggregation_view'))
    worker.test_run()

    assert sync_db.fetchval('select count(*) from messages') == 6
    r = cli.get(modify_url('/messages/email-test/aggregation/', settings, 'user-aggs'))
    assert r.status_code == 200, r.text
    data = r.json()
    histogram = data.pop('histogram')
    assert data == {
        'all_90_day': 5,
        'open_90_day': 1,
        'all_7_day': 5,
        'open_7_day': 1,
        'all_28_day': 5,
        'open_28_day': 1,
    }
    assert sorted(histogram, key=itemgetter('count')) == [
        {'count': 1, 'day': f'{date.today():%Y-%m-%d}', 'status': 'Opened'},
        {'count': 4, 'day': f'{date.today():%Y-%m-%d}', 'status': 'Sent'},
    ]


def test_user_aggregate_no_data(cli, settings, sync_db: SyncDb):
    sync_db.execute('insert into companies (code) values ($1)', 'testing')
    r = cli.get(modify_url('/messages/email-test/aggregation/', settings, 'testing'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data == {
        'histogram': [],
        'all_90_day': 0,
        'open_90_day': 0,
        'all_7_day': 0,
        'open_7_day': 0,
        'all_28_day': 0,
        'open_28_day': 0,
    }


def test_user_tags(cli, settings, send_email):
    uid1 = str(uuid.uuid4())
    send_email(
        uid=uid1,
        company_code='tagtest',
        tags=['trigger:broadcast', 'broadcast:123'],
        recipients=[
            {'address': '1@t.com', 'tags': ['user:1', 'shoesize:10']},
            {'address': '2@t.com', 'tags': ['user:2', 'shoesize:8']},
        ],
    )
    uid2 = str(uuid.uuid4())
    send_email(
        uid=uid2,
        company_code='tagtest',
        tags=['trigger:other'],
        recipients=[
            {'address': '3@t.com', 'tags': ['user:3', 'shoesize:10']},
            {'address': '4@t.com', 'tags': ['user:4', 'shoesize:8']},
        ],
    )

    send_email(uid=str(uuid.uuid4()), company_code='different1')
    send_email(uid=str(uuid.uuid4()), company_code='different2')

    r = cli.get(modify_url('/messages/email-test/', settings, 'tagtest') + '&tags=broadcast:123')
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 2, json.dumps(data, indent=2)
    assert {h['external_id'] for h in data['items']} == {f'{uid1}-1tcom', f'{uid1}-2tcom'}

    r = cli.get(modify_url('/messages/email-test/', settings, 'tagtest') + '&tags=user:2')
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1, json.dumps(data, indent=2)
    assert data['items'][0]['external_id'] == f'{uid1}-2tcom'

    r = cli.get(modify_url('/messages/email-test/', settings, 'tagtest') + '&tags=trigger:other&tags=shoesize:8')
    assert r.status_code == 200, r.text
    data = r.json()
    assert data['count'] == 1
    assert data['items'][0]['external_id'] == f'{uid2}-4tcom'


def test_search_emails(cli: TestClient, send_email, settings):
    send_email(recipients=[{'address': 'bon@jovi.com'}, {'address': 'robbie@willams.co.uk'}])
    r = cli.get(modify_url('/messages/email-test/', settings))
    assert r.json()['count'] == 2
    r = cli.get(modify_url(f'/messages/email-test/?{urlencode({"q": "bon@jovi.com"})}', settings))
    assert r.json()['count'] == 1
    r = cli.get(modify_url(f'/messages/email-test/?{urlencode({"q": "bon & jovi"})}', settings))
    assert r.json()['count'] == 0


def test_message_details(cli, settings, send_email, sync_db: SyncDb, worker, loop):
    msg_ext_id = send_email(company_code='test-details')

    data = {'ts': int(1e10), 'event': 'open', '_id': msg_ext_id, 'user_agent': 'testincalls'}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    assert worker.test_run() == 2

    message_id = sync_db.fetchval_b('select id from messages where :where', where=V('external_id') == msg_ext_id)
    r = cli.get(modify_url(f'/messages/email-test/{message_id}/', settings, 'test-details'))
    assert r.status_code == 200, r.text
    data = r.json()
    data['events'][0].pop('details')
    assert data == {
        'id': message_id,
        'external_id': msg_ext_id,
        'to_ext_link': None,
        'to_address': 'foobar@testing.com',
        'to_dst': '<foobar@testing.com>',
        'to_name': ' ',
        'send_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'subject': 'test message',
        'update_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'status': 'Opened',
        'method': 'email-test',
        'body': '<body>\nthis is a test\n</body>',
        'events': [{'status': 'Opened', 'datetime': RegexStr(r'\d{4}-\d{2}-\d{2}.*')}],
        'attachments': [],
        'cost': 0,
    }


def test_message_details_links(cli, settings, send_email, sync_db: SyncDb, worker, loop):
    msg_ext_id = send_email(
        company_code='test-details',
        recipients=[
            {
                'first_name': 'Foo',
                'last_name': 'Bar',
                'user_link': '/whatever/123/',
                'address': 'foobar@testing.com',
                'pdf_attachments': [
                    {'name': 'testing.pdf', 'html': '<h1>testing</h1>', 'id': 123},
                    {'name': 'different.pdf', 'html': '<h1>different</h1>'},
                ],
            }
        ],
    )
    message_id = sync_db.fetchval_b('select id from messages where :where', where=V('external_id') == msg_ext_id)
    data = {'ts': int(2e12), 'event': 'open', '_id': msg_ext_id, 'user_agent': 'testincalls'}
    r = cli.post('/webhook/test/', json=data)
    assert r.status_code == 200, r.text
    assert worker.test_run() == 2
    r = cli.get(modify_url(f'/messages/email-test/{message_id}/', settings, 'test-details'))
    assert r.status_code == 200, r.text
    data = r.json()
    data['events'][0].pop('details')
    assert data == {
        'id': message_id,
        'external_id': msg_ext_id,
        'to_ext_link': '/whatever/123/',
        'to_address': 'foobar@testing.com',
        'to_dst': 'Foo Bar <foobar@testing.com>',
        'to_name': 'Foo Bar',
        'send_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'subject': 'test message',
        'update_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
        'status': 'Opened',
        'body': '<body>\nthis is a test\n</body>',
        'method': 'email-test',
        'events': [{'status': 'Opened', 'datetime': RegexStr(r'\d{4}-\d{2}-\d{2}.*')}],
        'attachments': [['/attachment-doc/123/', 'testing.pdf'], ['#', 'different.pdf']],
        'cost': 0,
    }


def test_no_event_data(cli, settings, send_email, sync_db: SyncDb):
    msg_ext_id = send_email(
        company_code='test-details', recipients=[{'first_name': 'Foo', 'address': 'foobar@testing.com'}]
    )
    message_id = sync_db.fetchval_b('select id from messages where :where', where=V('external_id') == msg_ext_id)
    sync_db.executemany_b(
        'insert into events (:values__names) values :values',
        [
            Values(
                ts=(datetime(2032, 6, 1) + timedelta(days=i, hours=i * 2)).replace(tzinfo=timezone.utc),
                message_id=message_id,
                status=MessageStatus.send,
            )
            for i in range(3)
        ],
    )
    r = cli.get(modify_url(f'/messages/email-test/{message_id}/', settings, 'test-details'))
    assert r.json()['events'] == [
        {'status': 'Sent', 'datetime': '2032-06-01T00:00:00+00:00'},
        {'status': 'Sent', 'datetime': '2032-06-02T02:00:00+00:00'},
        {'status': 'Sent', 'datetime': '2032-06-03T04:00:00+00:00'},
    ]


def test_invalid_message_id(cli, sync_db: SyncDb, settings, send_email):
    msg_ext_id = send_email(
        company_code='test-details', recipients=[{'first_name': 'Foo', 'address': 'foobar@testing.com'}]
    )
    message_id = sync_db.fetchval_b('select id from messages where :where', where=V('external_id') == msg_ext_id)
    r = cli.get(modify_url(f'/messages/email-test/{message_id}/', settings, 'not_real_company'))
    assert r.status_code == 404

    r = cli.get(modify_url(f'/messages/email-test/999{message_id}/', settings, 'test-details'))
    assert r.status_code == 404


def test_many_events(cli, settings, send_email, sync_db: SyncDb):
    msg_ext_id = send_email(
        company_code='test-details', recipients=[{'first_name': 'Foo', 'address': 'foobar@testing.com'}]
    )
    message_id = sync_db.fetchval_b('select id from messages where :where', where=V('external_id') == msg_ext_id)
    sync_db.executemany_b(
        'insert into events (:values__names) values :values',
        [
            Values(
                ts=(datetime(2032, 6, 1) + timedelta(days=i, hours=i * 2)).replace(tzinfo=timezone.utc),
                message_id=message_id,
                status=MessageStatus.send,
                extra=json.dumps({'foo': 'bar', 'v': i}),
            )
            for i in range(55)
        ],
    )

    r = cli.get(modify_url(f'/messages/email-test/{message_id}/', settings, 'test-details'))
    assert r.status_code == 200, r.text
    events = r.json()['events']
    assert len(events) == 51
    assert sync_db.fetchval('select count(*) from events') == 55
    assert events[-1]['status'] == '5 more'


def test_user_sms_list(cli, settings, send_sms, sync_db: SyncDb):
    ext_id = send_sms(company_code='snapcrap')

    send_sms(uid=str(uuid.uuid4()), company_code='flip')
    r = cli.get(modify_url('/messages/sms-test/', settings, 'snapcrap'))
    assert r.status_code == 200, r.text
    data = r.json()
    assert data == {
        'items': [
            {
                'id': 1,
                'external_id': ext_id,
                'to_ext_link': None,
                'to_address': '+44 7896 541236',
                'to_dst': '<+44 7896 541236>',
                'to_name': ' ',
                'subject': 'this is a test apples',
                'send_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
                'update_ts': RegexStr(r'\d{4}-\d{2}-\d{2}.*'),
                'status': 'Sent',
                'method': 'sms-test',
                'cost': 0.012,
            },
        ],
        'count': 1,
        'spend': 0.012,
    }


def test_valid_signature(cli, settings, sync_db: SyncDb):
    sync_db.execute('insert into companies (code) values ($1)', 'whatever')
    args = dict(company='whatever', expires=round(datetime(2032, 1, 1).timestamp()))
    body = '{company}:{expires}'.format(**args).encode()
    args['signature'] = hmac.new(settings.user_auth_key, body, hashlib.sha256).hexdigest()
    r = cli.get('/messages/email-test/?' + urlencode(args))
    assert r.status_code == 200, r.text


def test_invalid_signature(cli, settings):
    args = dict(company='whatever', expires=round(datetime(2032, 1, 1).timestamp()))
    body = '{company}:{expires}'.format(**args).encode()
    args['signature'] = hmac.new(settings.user_auth_key, body, hashlib.sha256).hexdigest() + 'xxx'
    r = cli.get('/messages/email-test/?' + urlencode(args))
    assert r.status_code == 403, r.text
    assert {'message': 'Invalid token'} == r.json()


def test_invalid_expiry(cli, settings):
    args = dict(company='whatever', expires='xxx')
    body = '{company}:{expires}'.format(**args).encode()
    args['signature'] = hmac.new(settings.user_auth_key, body, hashlib.sha256).hexdigest()
    r = cli.get('/messages/email-test/?' + urlencode(args))
    assert r.status_code == 422, r.text
    assert {
        "detail": [
            {"loc": ["query", "expires"], "msg": "invalid datetime format", "type": "value_error.datetime"},
            {"loc": ["query", "expires"], "msg": "invalid datetime format", "type": "value_error.datetime"},
        ]
    } == r.json()


def test_sig_expired(cli, settings):
    args = dict(company='whatever', expires=round(datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp()))
    body = '{company}:{expires}'.format(**args).encode()
    args['signature'] = hmac.new(settings.user_auth_key, body, hashlib.sha256).hexdigest()
    r = cli.get('/messages/email-test/?' + urlencode(args))
    assert r.status_code == 403, r.text
    assert {'message': 'Token expired'} == r.json()



import base64
import pytest
from foxglove.db.helpers import SyncDb
from foxglove.test_server import DummyServer
from foxglove.testing import Client
from starlette.testclient import TestClient

from src.ext import ApiError, ApiSession
from tests.test_user_display import modify_url


def test_index(cli: TestClient):
    r = cli.get('/')
    assert r.status_code == 200
    assert 'Morpheus - The Greek God' in r.content.decode()


def test_index_head(cli: TestClient):
    r = cli.head('/')
    assert r.status_code == 200
    assert '' == r.text


def test_robots(cli: TestClient):
    r = cli.get('/robots.txt')
    assert r.status_code == 200
    assert 'User-agent: *' in r.text


def test_favicon(cli: TestClient):
    r = cli.get('/favicon.ico', allow_redirects=False)
    assert r.status_code == 200
    assert 'image' in r.headers['Content-Type']  # value can vary


def test_405(cli: TestClient):
    r = cli.post('/')
    assert r.status_code == 405, r.text


def test_create_subaccount_new_few_sent(cli: Client, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'foobar'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert r.json() == {'message': 'subaccount created'}
    assert dummy_server.log == ['POST /mandrill/subaccounts/add.json > 200']

    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {
        'message': 'subaccount already exists with only 42 emails sent, reuse of subaccount id permitted'
    }
    assert dummy_server.log == [
        'POST /mandrill/subaccounts/add.json > 200',
        'POST /mandrill/subaccounts/add.json > 500',
        'GET /mandrill/subaccounts/info.json > 200',
    ]


def test_create_subaccount_lots(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'lots-sent'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text

    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 409, r.text
    assert r.json() == {
        'message': 'subaccount already exists with 200 emails sent, reuse of subaccount id not permitted'
    }
    assert dummy_server.log == [
        'POST /mandrill/subaccounts/add.json > 200',
        'POST /mandrill/subaccounts/add.json > 500',
        'GET /mandrill/subaccounts/info.json > 200',
    ]


def test_create_subaccount_wrong_response(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'broken'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 400, r.text

    assert dummy_server.log == ['POST /mandrill/subaccounts/add.json > 500']


def test_create_subaccount_other_method(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    r = cli.post('/create-subaccount/email-test/', headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'no subaccount creation required for "email-test"'}

    assert dummy_server.log == []


def test_create_subaccount_invalid_key(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'foobar'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-keyX'})
    assert r.status_code == 403, r.text


def test_create_subaccount_on_send_email(cli: TestClient, sync_db: SyncDb, dummy_server, send_email):
    data = {'company_code': 'foobar'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert r.json() == {'message': 'subaccount created'}
    assert dummy_server.log == ['POST /mandrill/subaccounts/add.json > 200']

    assert sync_db.fetchval('select count(*) from companies') == 0

    send_email(company_code='foobar')
    assert sync_db.fetchval('select count(*) from companies') == 1


def test_create_subaccount_on_send_sms(cli: TestClient, sync_db: SyncDb, dummy_server, send_sms):
    data = {'company_code': 'foobar'}
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text
    assert r.json() == {'message': 'subaccount created'}
    assert dummy_server.log == ['POST /mandrill/subaccounts/add.json > 200']
    assert sync_db.fetchval('select count(*) from companies') == 0

    send_sms(company_code='foobar')
    assert sync_db.fetchval('select count(*) from companies') == 1


def test_user_list_subaccount_doesnt_exist(cli, settings, sync_db: SyncDb, dummy_server: DummyServer):
    r = cli.get(modify_url('/user/email-test/messages.json', settings))
    assert r.status_code == 404


def _create_test_subaccount(cli, data):
    r = cli.post('/create-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 201, r.text


def test_delete_subaccount(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'foobar'}
    _create_test_subaccount(cli, data)

    r = cli.post('/delete-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'deleted_messages=0 deleted_message_groups=0'}
    assert dummy_server.log == [
        'POST /mandrill/subaccounts/add.json > 200',
        'POST /mandrill/subaccounts/delete.json > 200',
    ]

    r = cli.post('/delete-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 404, r.text
    assert r.json() == {'message': "No subaccount exists with the id 'foobar'"}
    assert dummy_server.log == [
        'POST /mandrill/subaccounts/add.json > 200',
        'POST /mandrill/subaccounts/delete.json > 200',
        'POST /mandrill/subaccounts/delete.json > 500',
    ]


def test_delete_subaccount_multiple_branches(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'foobar'}
    sync_db.execute('insert into companies (code) values ($1)', 'foobar:1')
    sync_db.execute('insert into companies (code) values ($1)', 'foobar:2')
    sync_db.execute('insert into companies (code) values ($1)', 'notbar:1')

    r = cli.post('/delete-subaccount/email-test/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'deleted_messages=0 deleted_message_groups=0'}
    assert sync_db.fetchval('select count(*) from companies') == 1


def test_delete_subaccount_wrong_response(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    data = {'company_code': 'broken1'}
    _create_test_subaccount(cli, data)

    r = cli.post('/delete-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 400, r.text
    assert dummy_server.log == [
        'POST /mandrill/subaccounts/add.json > 200',
        'POST /mandrill/subaccounts/delete.json > 500',
    ]


def test_delete_subaccount_other_method(cli: TestClient, sync_db: SyncDb, dummy_server: DummyServer):
    r = cli.post(
        '/delete-subaccount/email-test/', json={'company_code': 'foobar'}, headers={'Authorization': 'testing-key'}
    )
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'deleted_messages=0 deleted_message_groups=0'}

    assert dummy_server.log == []


def test_delete_subaccount_invalid_key(cli: TestClient, sync_db: SyncDb):
    data = {'company_code': 'foobar'}
    r = cli.post('/delete-subaccount/email-mandrill/', json=data, headers={'Authorization': 'testing-keyX'})
    assert r.status_code == 403, r.text


def test_delete_subaccount_and_saved_messages(
    cli: TestClient, sync_db: SyncDb, send_email, send_sms, dummy_server: DummyServer
):
    send_email(company_code='foobar1')
    send_sms(company_code='foobar1')
    send_email(company_code='foobar2', recipients=[{'address': f'{i}@test.com'} for i in range(5)])
    assert sync_db.fetchval('select count(*) from companies') == 2
    assert sync_db.fetchval('select count(*) from message_groups') == 3
    assert sync_db.fetchval('select count(*) from messages') == 7

    fb1_data = {'company_code': 'foobar1'}
    _create_test_subaccount(cli, fb1_data)
    fb2_data = {'company_code': 'foobar2'}
    _create_test_subaccount(cli, fb2_data)

    r = cli.post('/delete-subaccount/email-mandrill/', json=fb1_data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'deleted_messages=2 deleted_message_groups=2'}

    assert sync_db.fetchval('select count(*) from message_groups') == 1
    assert sync_db.fetchval('select count(*) from messages') == 5

    r = cli.post('/delete-subaccount/email-mandrill/', json=fb2_data, headers={'Authorization': 'testing-key'})
    assert r.status_code == 200, r.text
    assert r.json() == {'message': 'deleted_messages=5 deleted_message_groups=1'}

    assert sync_db.fetchval('select count(*) from message_groups') == 0
    assert sync_db.fetchval('select count(*) from messages') == 0

    send_email(company_code='foobar3')
    assert sync_db.fetchval('select count(*) from message_groups') == 1
    assert sync_db.fetchval('select count(*) from messages') == 1

    _create_test_subaccount(cli, {'company_code': 'foobar3'})
    with pytest.raises(TypeError):
        cli.post(
            '/delete-subaccount/email-mandrill/',
            json={'company_code': object()},
            headers={'Authorization': 'testing-key'},
        )
    assert sync_db.fetchval('select count(*) from message_groups') == 1
    assert sync_db.fetchval('select count(*) from messages') == 1


def test_missing_link(cli: TestClient):
    r = cli.get('/lxxx')
    assert r.status_code == 404, r.text
    assert (
        '<p>404: No redirect could be found for "http://testserver/lxxx", this link may have expired.</p>'
    ) in r.text


def test_missing_url_with_arg(cli: TestClient):
    url = 'https://example.com/foobar'
    r = cli.get('/lxxx?u=' + base64.urlsafe_b64encode(url.encode()).decode(), allow_redirects=False)
    assert r.status_code == 307, r.text
    assert r.headers['Location'] == url


def test_missing_url_with_arg_bad(cli: TestClient):
    r = cli.get('/lxxx?u=xxx', allow_redirects=False)
    assert r.status_code == 404, r.text


def test_api_error(settings, loop, dummy_server: DummyServer):
    s = ApiSession(dummy_server.server_name, settings)
    with pytest.raises(ApiError) as exc_info:
        loop.run_until_complete(s.get('/foobar'))
    assert str(exc_info.value) == f'GET {dummy_server.server_name}/foobar, unexpected response 404'


def test_settings(settings):
    assert settings.pg_host == 'localhost'
    assert settings.pg_port == 5432
    assert settings.pg_name == 'morpheus_test'



from setuptools import setup

from src.version import VERSION

setup(
    name='morpheus-mail',
    version=str(VERSION),
    description='Email rendering engine from morpheus',
    long_description="""
Note: this only installs the rendering logic for `morpheus <https://github.com/tutorcruncher/morpheus>` \
for testing and email preview.

Everything else is excluded to avoid installing unnecessary packages.
""",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Topic :: Internet',
    ],
    author='Samuel Colvin',
    author_email='s@muelcolvin.com',
    url='https://github.com/tutorcruncher/morpheus',
    license='MIT',
    packages=['src.render'],
    package_dir={'src.render': 'src/render'},
    python_requires='>=3.6',
    zip_safe=True,
    install_requires=[
        'chevron>=0.11.1',
        'libsass>=0.13.2',
        'misaka>=2.1.1',
    ],
)
